#include "stdafx.h"
#include "ilmatrixoperations.h"

//************************************************************************
// ilEuclidDist
//   Computes Euclidean distance between sets of points in src1 and src2.
//   src1 is m1*n array; src2 is m2*n array where m1 and m2 are
//   the number of points in each set and n is the dimensionality
//   of the Euclidean space. The result dst is an m1*m2 matrix
//   with distances between all pairs of points.
//   If m1=m2=1 Euclidean distance between two vectors is computed.
//************************************************************************
void ilEuclidDist(const CvArr *src1arr, const CvArr *src2arr, CvArr *dstarr){
  int t1,t2; double dt;
  CvMat *src1=(CvMat*)src1arr;
  CvMat *src2=(CvMat*)src2arr;
  CvMat *dst=(CvMat*)dstarr;
  int m1=src1->rows, m2=src2->rows, n=src1->cols, type=src1->type;

#if 0 
  //*** matrix version
  CvMat *ones_nm2 = cvCreateMat(n,m2,type);
  CvMat *ones_m1n = cvCreateMat(m1,n,type);
  CvMat *src1sq = cvCreateMat(m1,n,type);
  CvMat *src2sq = cvCreateMat(m2,n,type);
  CvMat *res1 = cvCreateMat(m1,m2,type);
  CvMat *res2 = cvCreateMat(m1,m2,type);

  t1=cvGetTickCount();
  cvMul(src1,src1,src1sq);
  cvMul(src2,src2,src2sq);
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);
  cvSet(ones_nm2,cvScalar(1.0));
  cvSet(ones_m1n,cvScalar(1.0));

  //*** Euclidean distance with matrix multiplications
  t1=cvGetTickCount();
  cvMatMul(src1sq,ones_nm2,res1);
  cvGEMM(ones_m1n,src2sq,1.0,res1,1,res2,CV_GEMM_B_T);
  cvGEMM(src1,src2,-2.0,res2,1,dst,CV_GEMM_B_T);
  cvPow(dst,dst,.5);
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

  cvReleaseMat(&ones_nm2);
  cvReleaseMat(&ones_m1n);
  cvReleaseMat(&src1sq);
  cvReleaseMat(&src2sq);
  cvReleaseMat(&res1);
  cvReleaseMat(&res2);
#endif

#if 1  
  //*** element-wise version
  CvMat *src1sq = cvCreateMat(m1,n,type);
  CvMat *src2sq = cvCreateMat(m2,n,type);
  cvMul(src1,src1,src1sq);
  cvMul(src2,src2,src2sq);

  float *ptrsrc1sq, *ptrsrc2sq, *ptrdst;
  float v0, v1, v2, v3, v4, v5, v6, v7, v8, v9, vdst;
  ptrsrc1sq=(float*)cvPtr2D(src1sq,0,0);
  ptrsrc2sq=(float*)cvPtr2D(src2sq,0,0);
  ptrdst =(float*)cvPtr2D(dst,0,0);
  int indsrc1sq, indsrc2sq, inddst=0;
  for (int i=0; i<m1; i++){
	indsrc1sq=i*n;
	v0=ptrsrc1sq[indsrc1sq];   v1=ptrsrc1sq[indsrc1sq+1];
	v2=ptrsrc1sq[indsrc1sq+2]; v3=ptrsrc1sq[indsrc1sq+3];
	v4=ptrsrc1sq[indsrc1sq+4]; v5=ptrsrc1sq[indsrc1sq+5];
	v6=ptrsrc1sq[indsrc1sq+6]; v7=ptrsrc1sq[indsrc1sq+7];
	v8=ptrsrc1sq[indsrc1sq+8]; v9=ptrsrc1sq[indsrc1sq+9];
	for (int j=0; j<m2; j++){
	  inddst=i*m2+j;
	  indsrc2sq=j*n;
	  vdst=(v0-ptrsrc2sq[indsrc2sq])*(v0-ptrsrc2sq[indsrc2sq]);
	  vdst+=(v1-ptrsrc2sq[indsrc2sq+1])*(v1-ptrsrc2sq[indsrc2sq+1]);
	  vdst+=(v2-ptrsrc2sq[indsrc2sq+2])*(v2-ptrsrc2sq[indsrc2sq+2]);
	  vdst+=(v3-ptrsrc2sq[indsrc2sq+3])*(v3-ptrsrc2sq[indsrc2sq+3]);
	  vdst+=(v4-ptrsrc2sq[indsrc2sq+4])*(v4-ptrsrc2sq[indsrc2sq+4]);
	  vdst+=(v5-ptrsrc2sq[indsrc2sq+5])*(v5-ptrsrc2sq[indsrc2sq+5]);
	  vdst+=(v6-ptrsrc2sq[indsrc2sq+6])*(v6-ptrsrc2sq[indsrc2sq+6]);
	  vdst+=(v7-ptrsrc2sq[indsrc2sq+7])*(v7-ptrsrc2sq[indsrc2sq+7]);
	  vdst+=(v8-ptrsrc2sq[indsrc2sq+8])*(v8-ptrsrc2sq[indsrc2sq+8]);
	  vdst+=(v9-ptrsrc2sq[indsrc2sq+9])*(v9-ptrsrc2sq[indsrc2sq+9]);
	  ptrdst[inddst]=vdst;
	}
  }
  cvReleaseMat(&src1sq);
  cvReleaseMat(&src2sq);
#endif
}

//*******************************************************************
// ilEuclidNorm
//   normalizes vectors of dim=n in mxn input matrix with the
//   Euclidean norm
//*******************************************************************
void ilEuclidNorm(const CvArr *srcarr, CvArr *dstarr){
  CvMat *src=(CvMat*)srcarr;
  CvMat *dst=(CvMat*)dstarr;

  int m=src->rows, n=src->cols, type=src->type;
  CvMat *ones = cvCreateMat(n,1,type);
  CvMat *res1 = cvCreateMat(m,n,type);
  CvMat *norm = cvCreateMat(m,1,type);
  cvSet(ones,cvScalar(1.0));

  //*** compute the norm
  cvMul(src,src,res1);
  cvMatMul(res1,ones,norm);
  cvPow(norm,norm,.5);

  //*** normalize columns of src
#if 0 //*** matrix version
  CvMat* normmat = cvCreateMat(m,n,CV_32FC1);
  CvMat* onesrow = cvCreateMat(1,n,CV_32FC1);
  cvSet(onesrow,cvScalar(1.0));
  cvMatMul(norm,onesrow,normmat);
  cvDiv(src,normmat,dst);
  cvReleaseMat(&normmat);
  cvReleaseMat(&onesrow);
#endif

#if 1 // *** column version
  CvMat colmat1 = cvMat(0,0,0,0);
  CvMat colmat2 = cvMat(0,0,0,0);
  for (int i=0; i<n; i++){
    cvGetCol(src,&colmat1,i);
    cvGetCol(dst,&colmat2,i);
    cvDiv(&colmat1,norm,&colmat2);
  }
#endif

  cvReleaseMat(&ones);
  cvReleaseMat(&res1);
  cvReleaseMat(&norm);
}

//*********************************************************************
// ilMaxRowFloat
//   Finds max values and corresponding indicies along the rows
//   of an input 2-dim. float mxn matrix. maxvalarr and maxindarr
//   should have size mx1. maxindarr should be of type short.
//*********************************************************************
void ilMaxRowFloat(const CvArr *srcarr, CvArr *maxvalarr, CvArr *maxindarr, int absflag){
  CvMat *src=(CvMat*)srcarr;
  CvMat *maxvalmat=(CvMat*)maxvalarr;
  CvMat *maxindmat=(CvMat*)maxindarr;

  short maxind;
  float maxval,val;
  int m=src->rows, n=src->cols, type=src->type, indsrc;
  float *ptrsrc=(float*)cvPtr2D(src,0,0);
  float *ptrmaxval=(float*)cvPtr2D(maxvalmat,0,0);
  short *ptrmaxind=(short*)cvPtr2D(maxindmat,0,0);
  for (int i=0; i<m; i++){
    indsrc=i*n;
	maxval=ptrsrc[indsrc];
	maxind=0;
	for (int j=0; j<n; j++){
	  val=ptrsrc[indsrc+j];
	  if (absflag) {
	    if (val<0) val=-val;
	  }
	  if (val>maxval){
	    maxind=j;
		maxval=val;
      }
	}
	ptrmaxval[i]=maxval;
	ptrmaxind[i]=maxind;
  }
}

//*********************************************************************
// ilMaxColFloat
//   Finds max values and corresponding indicies along the columns
//   of an input 2-dim. float mxn matrix. maxvalarr and maxindarr
//   should have size nx1. maxindarr should be of type short.
//*********************************************************************
void ilMaxColFloat(const CvArr *srcarr, CvArr *maxvalarr, CvArr *maxindarr, int absflag){
  CvMat *src=(CvMat*)srcarr;
  CvMat *maxvalmat=(CvMat*)maxvalarr;
  CvMat *maxindmat=(CvMat*)maxindarr;

  short maxind;
  float maxval,val;
  int m=src->rows, n=src->cols, type=src->type, indsrc;
  float *ptrsrc=(float*)cvPtr2D(src,0,0);
  float *ptrmaxval=(float*)cvPtr2D(maxvalmat,0,0);
  short *ptrmaxind=(short*)cvPtr2D(maxindmat,0,0);
  for (int i=0; i<n; i++){
    indsrc=i;
	maxval=ptrsrc[indsrc];
	if (absflag) {
	    if (maxval<0) maxval=-maxval;
	}	
	maxind=0;
	for (int j=0; j<m; j++){
	  val=ptrsrc[indsrc+j*n];
	  if (absflag) {
	    if (val<0) val=-val;
	  }
	  if (val>maxval){
	    maxind=j;
		maxval=val;
      }
	}
    ptrmaxval[i]=maxval;
    ptrmaxind[i]=maxind;
  }
}

//*********************************************************************
// ilMaxProjectFloat
//   Projects columns of mxd matrix stcarr2 onto rows of nxm matrix srcarr1
//   and finds max values and corresponding indicies along the columns
//   of a resulting nxd projection matrix. Output matrices maxvalarr and maxindarr
//   should have size dx1. maxindarr should be of type short.
//*********************************************************************
void ilMaxProjectFloat(const CvArr *srcarr1, const CvArr *srcarr2, CvArr *maxvalarr, CvArr *maxindarr){

  CvMat *src1=(CvMat*)srcarr1;
  CvMat *src2=(CvMat*)srcarr2;
  CvMat *maxvalmat=(CvMat*)maxvalarr;
  CvMat *maxindmat=(CvMat*)maxindarr;

  short maxind;
  float maxval, val;
  int m=src1->rows, n=src1->cols, d=src2->cols, type=src1->type, indsrc1, indsrc2;
  float *ptrsrc1=(float*)cvPtr2D(src1,0,0);
  float *ptrsrc2=(float*)cvPtr2D(src2,0,0);
  float *ptrmaxval=(float*)cvPtr2D(maxvalmat,0,0);
  short *ptrmaxind=(short*)cvPtr2D(maxindmat,0,0);
  float v0,v1,v2,v3,v4,v5,v6,v7,v8,v9;

  for (int i=0; i<d; i++){
    indsrc2=i;
	maxval=0;
	for (int j=0; j<m; j++){
	  indsrc1=j*n;
	  val=0.0;
	  for (int k=0; k<n; k++){
        val+=ptrsrc2[indsrc2+k*d]*ptrsrc1[indsrc1+k];
	  }
   	  if ((val>maxval)||(j==0)){
	    maxind=j;
        maxval=val;
	  }
	}
	ptrmaxval[i]=maxval;
	ptrmaxind[i]=maxind;
  }
}

//**************************************************************
// ilMatTestFun
//   performs tests of matrix functions:
//   ilEuclidDist, ilEuclidNorm, ilMinAtDim
//***************************************************************
void ilMatFunTest(void){

#if 0 //*** testing ilEuclidNorm (comparison with matlab)
	char matfnamein[] = "C:/temp/matdata3.xml";
	char matfnameout[] = "C:/temp/matdata4.xml";
	CvFileStorage* fsr=cvOpenFileStorage(matfnamein,NULL,CV_STORAGE_READ);
	CvFileNode* fn=cvGetRootFileNode(fsr,0);
	CvFileNode* mn1=cvGetFileNodeByName(fsr,0,"matrix1");
	CvMat *mat1=(CvMat*)(cvRead(fsr,mn1,NULL));
    cvReleaseFileStorage(&fsr);

    //*** normalize
	CvMat *mat2 = cvCreateMat(mat1->rows,mat1->cols,mat1->type);
    ilEuclidNorm(mat1,mat2);

	//*** Save result in xml
	CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
    cvWrite(fsw,"matrix2",mat2,cvAttrList(0,0));
    cvReleaseFileStorage( &fsw );

	cvReleaseMat(&mat1);
	cvReleaseMat(&mat2);
	
	/*
	% Matlab test code
	m1=rand(10,5);
	xmlsavematrices('C:/temp/matdata3.xml',{m1},{'matrix1'});
    %%% run ilMatFunTest();
	[matval,matname]=xmlreadmatrices('c:/temp/matdata4.xml');
	enorm=euclidnorm(m1);
	err=mean(abs(enorm(:)-matval{1}(:)))
	*/
#endif

#if 0 //*** testing ilEuclidDist (comparison with matlab)
	char matfnamein[] = "C:/temp/matdata1.xml";
	char matfnameout[] = "C:/temp/matdata2.xml";
	CvFileStorage* fsr=cvOpenFileStorage(matfnamein,NULL,CV_STORAGE_READ);
	CvFileNode* fn=cvGetRootFileNode(fsr,0);
	CvFileNode* mn1=cvGetFileNodeByName(fsr,0,"matrix1");
	CvMat *mat1=(CvMat*)(cvRead(fsr,mn1,NULL));
	CvFileNode* mn2=cvGetFileNodeByName(fsr,0,"matrix2");
	CvMat *mat2=(CvMat*)(cvRead(fsr,mn2,NULL));
    cvReleaseFileStorage(&fsr);
    
	//*** compute Euclidean distance
	CvMat *mat3 = cvCreateMat(mat1->rows,mat2->rows,mat1->type);
    ilEuclidDist(mat1,mat2,mat3);

	//*** Save result in xml
	CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
    cvWrite(fsw,"matrix3",mat3,cvAttrList(0,0));
    cvReleaseFileStorage( &fsw );

	cvReleaseMat(&mat1);
	cvReleaseMat(&mat2);
	cvReleaseMat(&mat3);

	/*
	% Matlab test code
	m1=rand(10,5); m2=rand(20,5);
	xmlsavematrices('C:/temp/matdata1.xml',{m1,m2},{'matrix1','matrix2'});
    %%% run ilMatFunTest();
	[matval,matname]=xmlreadmatrices('c:/temp/matdata2.xml');
	edist=eucliddist(m1,m2);
	err=mean(abs(edist(:)-matval{1}(:)))
	*/
#endif
}

void ilRGB2Gray(IplImage *src, CvArr *dstarr){
    CvMat* dst=(CvMat*)dstarr;
	
	uchar *srcptr=(uchar*)cvPtr2D(src,0,0);
	uchar *dstptr=(uchar*)cvPtr2D(dst,0,0);
	int w=src->width; int w2=w*2;
	for (int i=0; i<(src->width*src->height); i++)
        dstptr[i]=(uchar)(div((int)srcptr[3*i]+(int)srcptr[3*i+1]+(int)srcptr[3*i+2],3).quot);
        //dstptr[i]=(uchar)(div((int)srcptr[i]+(int)srcptr[i+w]+(int)srcptr[i+w2],3).quot);
}