#include "xmlloadmodel.h"

#include <libxml++/libxml++.h>
#include <iostream>
#include <boost/lexical_cast.hpp>


xmlpp::Node::NodeList get_node_list_by_name(xmlpp::Node const * node, std::string const & name)
{
    xmlpp::Node::NodeList result;
    xmlpp::Node::NodeList const list = node->get_children();
    for(xmlpp::Node::NodeList::const_iterator cit = list.begin();
                                              cit != list.end(); ++cit)
    {
        if ((*cit)->get_name() == name)
            result.push_back(*cit);
    }
    return result;
}

xmlpp::Node const * get_node_by_name(xmlpp::Node const * node, std::string const & name)
{
    xmlpp::Node::NodeList const list = get_node_list_by_name(node, name);
    assert( list.size() == 1 );
    return *list.begin();
}

template<typename T>
T get_from_node(xmlpp::Node const * node)
{
    xmlpp::TextNode const * n = dynamic_cast<xmlpp::TextNode const *>(node);
    if (!n)
    {
        xmlpp::Node::NodeList const list = node->get_children();
        assert( list.size() );
        n = dynamic_cast<xmlpp::TextNode const *>( *list.begin() );
    }
    assert(n);
    try
    {
        return boost::lexical_cast<T>(n->get_content());
    }
    catch(...)
    {
        std::cerr
            << "cannot convert [" << n->get_content() << "] to some desired type..."
            << typeid(T).name() << "\n" << std::flush;
        throw;
    }
}

template<typename T>
T get_from_node(xmlpp::Node const * node, std::string const & name)
{
    try
    {
        return get_from_node<T>(get_node_by_name(node, name));
    }
    catch(...)
    {
        std::cerr << "...error in get_from_node(*," << name << ")\n" << std::flush;
        throw;
    }
}

void do_load_model(xmlpp::Node const * node, Model & model)
{
    using namespace xmlpp;
    typedef xmlpp::Node::NodeList::const_iterator list_iterator;
    Node const * child = NULL;

    assert(node && node->get_name() == "model");

    model.clear();

    // read fun
    Node::NodeList smlist = get_node_list_by_name(node, "stagemodel");
    for(list_iterator smit = smlist.begin(); smit != smlist.end(); ++smit)
    {
        Stage stage;
        // read stagemodel
        xmlpp::Node const * smodelNode = (*smit);

        stage.stage_threshold = get_from_node<double>(smodelNode, "stage_threshold");
        stage.exitflag = get_from_node<int>(smodelNode, "exitflag");

        Node::NodeList list = get_node_list_by_name(smodelNode, "member");
        for(list_iterator it = list.begin(); it != list.end(); ++it)
        {
            StageRule rule;

            Node const * memberNode = *it;
            rule.Alpha = get_from_node<double>(memberNode, "Alpha");
            rule.Z = get_from_node<double>(memberNode, "Z");
            rule.WeightedErr = get_from_node<double>(memberNode, "WeightedErr");
            rule.ErrBound = get_from_node<double>(memberNode, "ErrBound");

            Node const * ruleNode = get_node_by_name(memberNode, "rule");
            rule.type = (StageRuleType)get_from_node<int>(ruleNode, "ftype");

            child = get_node_by_name(ruleNode, "rect");
            rule.left   = get_from_node<int>(child, "x");
            rule.top    = get_from_node<int>(child, "y");
            rule.width  = get_from_node<int>(child, "w");
            rule.height = get_from_node<int>(child, "h");
    
            rule.trainerror = get_from_node<double>(ruleNode, "trainerror");
    
            child = get_node_by_name(ruleNode, "sz");
//            rule.size.x = get_from_node<int>(child, "x");
//            rule.size.y = get_from_node<int>(child, "y");
//            rule.size.z = get_from_node<int>(child, "z");

            Node const * rmodelNode = get_node_by_name(ruleNode, "rule_model");

            Node::NodeList wlist = get_node_list_by_name(get_node_by_name(rmodelNode, "W"), "w");
            for(list_iterator wit = wlist.begin(); wit != wlist.end(); ++ wit)
            {
                rule.W.push_back( get_from_node<double>(*wit) );
            }
            rule.b = get_from_node<double>(rmodelNode, "b");
            rule.separab = get_from_node<double>(rmodelNode, "separab");
    
            stage.rules.push_back( rule );
        }

        model.push_back(stage);
    }
}

void xmlloadmodel(std::string const & file, Model & model)
{
    try
    {
        xmlpp::DomParser parser;
        parser.set_validate( false );
        parser.set_substitute_entities(); //We just want the text to be resolved/unescaped automatically.
        parser.parse_file(file);
        if(parser)
        {
            //Walk the tree:
            xmlpp::Node const * pNode = parser.get_document()->get_root_node(); //deleted by DomParser.
            do_load_model(pNode, model);
        }
    }
    catch(const std::exception& ex)
    {
        std::cout << "Exception caught: " << ex.what() << std::endl;
    }
}
